from pulp import (
    LpMaximize,
    LpMinimize,
    LpVariable,
    LpBinary,
    LpConstraintLE,
    LpConstraintGE,
    LpConstraint,
    LpAffineExpression,
)

from muoblp.model.multi_objective_lp import MultiObjectiveLpProblem


def get_constraint_sign(constraint: str) -> int:
    if "<=" in constraint:
        return constraint.index("<")
    if ">=" in constraint:
        return constraint.index(">")
    raise Exception("Unexpected constraint sign")


def read_lp_file(filename) -> MultiObjectiveLpProblem:
    problem_data = {
        "name": "",
        "sense": None,  # 'Minimize' or 'Maximize'
        "constraints": [],
        "variables": {},
        "objectives": [],
    }

    with open(filename, "r") as f:
        lines = f.readlines()

    problem_data["name"] = lines[0].split("*")[1].strip()
    problem_data["sense"] = (
        LpMaximize if lines[1].lower().startswith("maximize") else LpMinimize
    )

    for variable_line in lines[
        lines.index("Binaries\n") + 1 : lines.index("End\n")
    ]:
        variable = LpVariable(variable_line.strip(), cat=LpBinary)
        variable.setInitialValue(0)
        problem_data["variables"][variable_line.strip()] = variable

    for constraint_line in lines[
        lines.index("Subject To\n") + 1 : lines.index("Bounds\n")
    ]:
        name_right_idx = constraint_line.index(":")
        sign_idx = get_constraint_sign(constraint_line)
        c_name = constraint_line[:name_right_idx]
        c_lhs_str = constraint_line[name_right_idx + 2 : sign_idx]

        c_lhs = [
            (problem_data["variables"][var], int(coef))
            for coef_var in c_lhs_str.split("+")
            for coef, var in [coef_var.strip().split(" ")]
        ]

        c_sense = (
            LpConstraintLE
            if constraint_line[sign_idx : sign_idx + 2] == "<="
            else LpConstraintGE
        )
        c_rhs = int(constraint_line[sign_idx + 2 :].strip())
        constraint = LpConstraint.from_dict(
            {
                "coefficients": LpAffineExpression(c_lhs),
                "constant": -c_rhs,
                "name": c_name,
                "sense": c_sense,
                "pi": -0.0,  # TODO: What should be the value?
            }
        )
        problem_data["constraints"].append(constraint)

    for target_line in lines[
        lines.index("OBJECTIVES:\n") + 1 : lines.index("END_OBJECTIVES:\n")
    ]:
        name_right_idx = target_line.index(":")
        t_name = target_line[:name_right_idx]
        target = LpAffineExpression(
            [
                [problem_data["variables"][var.strip()], 1]
                for var in target_line[name_right_idx + 2 :].split("+")
            ],
            name=t_name,
        )
        problem_data["objectives"].append(target)

    problem = MultiObjectiveLpProblem(
        problem_data["name"], problem_data["sense"], problem_data["objectives"]
    )
    problem.addVariables(problem_data["variables"].values())
    for constraint in problem_data["constraints"]:
        problem.addConstraint(constraint)

    return problem



from typing import TypeVar, List

T = TypeVar("T")


def flatten(collection: List[List[T]]) -> List[T]:
    return [x for xs in collection for x in xs]


def get_geometric_ratios(numbers: List[float]) -> List[float]:
    ratios = []
    for i in range(len(numbers) - 1):
        ratios.append(numbers[i + 1] / numbers[i])
    return ratios



from typing import List

import pulp
from pulp import LpProblem, LpAffineExpression, LpMaximize, LpMinimize

from muoblp.utils.lpWriterUtils import expression_to_lp_format

# Override line limit for writeLP to output single line values
pulp.const.LpCplexLPLineSize = 100000


class MultiObjectiveLpProblem(LpProblem):
    def __init__(
        self,
        name: str,
        sense: LpMaximize | LpMinimize = LpMaximize,
        objectives: List[LpAffineExpression] = None,
    ) -> None:
        super().__init__(name, sense=sense)
        self._objectives = objectives

    @property
    def objectives(self) -> List[LpAffineExpression]:
        return self._objectives

    def setObjectives(self, objectives: List[LpAffineExpression]) -> None:
        self._objectives = objectives

    # TODO: Decide how to handle fixObjective and restoreObjective

    def writeLP(self, filename, writeSOS=1, mip=1, max_length=100):
        super().writeLP(filename, writeSOS, mip, max_length)
        with open(filename, "a", encoding="utf-8") as file:
            file.write("OBJECTIVES:\n")
            for objective in self.objectives:
                file.write(expression_to_lp_format(objective))
            file.write("END_OBJECTIVES:\n")
        return

    # TODO: override __iadd__ to append objective to the list of objectives



from pulp import LpAffineExpression


def expression_to_lp_format(
    expression: LpAffineExpression, constant: int = 1
) -> str:
    """
    copy-paste of `LpAffineExpression::asCplexLpAffineExpression` with line length limit modification
    returns a string that represents the Affine Expression in lp format
    """
    # refactored to use a list for speed in iron python
    result, line = expression_to_variables_only(expression)
    if not expression:
        term = f" {expression.constant}"
    else:
        term = ""
        if constant:
            if expression.constant < 0:
                term = " - %s" % (-expression.constant)
            elif expression.constant > 0:
                term = f" + {expression.constant}"
    line += [term]
    result += ["".join(line)]
    result = "%s\n" % "\n".join(result)
    return result


def expression_to_variables_only(expression: LpAffineExpression):
    """
    helper for expression_to_lp_format
    copy-paste of `LpAffineExpression::asCplexVariablesOnly` with line length limit modification
    """
    result = []
    line = [f"{expression.name}:"]
    notFirst = 0
    variables = expression.sorted_keys()
    for v in variables:
        val = expression[v]
        if val < 0:
            sign = " -"
            val = -val
        elif notFirst:
            sign = " +"
        else:
            sign = ""
        notFirst = 1
        if val == 1:
            term = f"{sign} {v.name}"
        else:
            # adding zero to val to remove instances of negative zero
            term = f"{sign} {val + 0:.12g} {v.name}"

        line += [term]
    return result, line
